; TODO:
; - remove image widget part which depends on PLT-Scheme
; - remove XML part which depends on PLT-Scheme
; - check the error mechanism


;/****m*
;* NAME
;*   dialog
;* DESCRIPTION
;*   This module can be used to define dialog windows. The dialog window
;*   is defined in an XML file and the actions are defined in a Scheme file.
;*   The concept of this system is similar to DCL (Dialog Control Language)
;*   developed by AutoCAD. 
;*
;*   Use the sx-load-dialog function to load an XML file describing the dialog
;*   window. It returns a list. Use the sx-new-dialog window to create the 
;*   objects in the dialog window. Then setup some parameters in the window
;*   with the sx-value-set, sx-label-set, sx-clear-list, sx-add-list,
;*   sx-image-fill, sx-image-file functions. Use the sx-action-set function
;*   to setup callback functions. In the callback function of the OK or CANCEL
;*   buttons use the sx-hide-dialog function to close down the window. To display
;*   the window use the sx-show-dialog function. If you want to destroy the
;*   objects use the sx-unload-dialog function.
;*
;*   Parameters of widgets
;*     dialog : key, label, width, height
;*     button : key, label, width, height
;*     toogle : key, label, width, height, value ("0", "1")
;*     message : key, label, width, height
;*     edit-box : key, label, width, height, value
;*     popup-list : key, label, width, height
;*     list-box : key, width, height
;*     image : key, width, height
;*     row : key, width, height, horizontal-alignment
;*     column : key, width, height, vertical-alignment
;*     boxed-column : key, label, width, height, vertical-alignment
;*     tab-panel : key, width, height
;*     tab : key, label
;*     spacer : key, width, height, horizontal, vertical
;*
;*  Example:
;*
;*    The XML file:
;*
;*      <dialog key="test" width="300" height="200">
;*        <tab-panel key="panel">
;*          <tab> key="files" label="Files to set">
;*            <edit-box key="edit-file" width="100" />
;*            <list-box key="list1" label="Loaded files" />
;*          </tab>
;*          <tab key="options" label="Options">
;*            <toggle key="black" label="Black?" />
;*            <toggle key="cnt" label="Contour?" />
;*          </tab>
;*        </tab-panel>
;*        <row key="ok-cancel">
;*          <button key="ok" label="OK" width="80" />
;*          <button key="cancel" label="Cancel" width="80" />
;*        </row>
;*      </dialog>
;*
;*    The program:
;*
;*      (require (file "dialog.scm"))
;*      (define xml (sx-load-dialog "dialog_1.xml"))
;*      (define win (sx-new-dialog xml))
;*      (define result #f)
;*      
;*      (sx-clear-list win "list1")
;*      (sx-add-list win "list1" "c:/dir1/file1")
;*      (sx-add-list win "list1" "c:/dir1/file2")
;*      (sx-add-list win "list1" "c:/dir2/file3")
;*      
;*      (sx-value-set win "edit-file" "Something")
;*      (sx-value-set win "black" #t)
;*      
;*      (sx-action-set win "cancel" (lambda () (sx-hide-dialog)))
;*      (sx-action-set win "ok" 
;*                     (lambda ()
;*                       ; check values and set result
;*                       (sx-hide-dialog)))
;*      
;*      (sx-show-dialog win)
;*      (sx-unload-dialog win)
;******/




(module dialog mzscheme

  (require (lib "xml.ss" "xml")
           ;;;
           (file "gui/sx-hash.scm")
           (file "gui/sx-dialog.scm")
           (file "gui/sx-button.scm")
           (file "gui/sx-check-box.scm")
           (file "gui/sx-text.scm")
           (file "gui/sx-entry.scm")
           (file "gui/sx-choice.scm")
           (file "gui/sx-canvas.scm")
           (file "gui/sx-list-box.scm")
           (file "gui/sx-container.scm")
  )
  
  (provide sx-load-dialog
           sx-new-dialog
           sx-show-dialog
           sx-hide-dialog
           sx-unload-dialog
           sx-action-set
           sx-image-fill
           sx-image-file
           sx-clear-list
           sx-add-list
           sx-label-set
           sx-value-set
           sx-value-get
  )  
  
  ; this hash table stores all loaded dialog windows
  (define table-dialog (sx-hash-make))
  
  ; the list of opened dialog windows
  (define opened-dialog '())
  
  
  
  ; FIXME: PLT dependent part for XML reading and parsing
  (collapse-whitespace #t)
  (xexpr-drop-empty-attributes #t)
  ; loads in a description of a dialog window in XML format
  ; FIXME: PLT dependent part for XML reading and parsing
  (define (sx-load-dialog filename)
    (let ((fp #f) (expr #f))
      (with-handlers ((exn:fail:filesystem? (lambda (exn) #f)))
        (set! fp (open-input-file filename))
      )
      (if (port? fp)
        (begin
          (with-handlers ((exn:xml? (lambda (exn) #f)))
            (set! expr (xml->xexpr (document-element (read-xml fp))))
          )
          (close-input-port fp)
        )
      )
      (cond
        ((not fp)   'error-file)
        ((not expr) 'error-xml)
        (else
         (let loop ((lst expr))
           (cond
             ((not (list? lst)) lst)
             ((null? lst) '())
             ((equal? (car lst) " ") (loop (cdr lst)))
             (else (cons (loop (car lst)) (loop (cdr lst))))
           )
         )
        )
      )
    )
  )
  
  
  ; parse a window layout from a list data structure
  ; returns a string if the window could be created
  ; if a component is not valid, it is not a fatal failure, 
  ;   it is simply not added to the window
  ; in the case of an error it returns a symbol
  (define (sx-new-dialog xml)
    (if (and (equal? (list-ref xml 0) 'dialog)
             (>= (length xml) 2))
      (let*
        ((props (list-ref xml 1))
         (key (assoc 'key props))
        )
        (if key
          (let
            ((label  (assoc 'label props))
             (width  (assoc 'width props))
             (height (assoc 'height props))
            )
            (let*
              ((label  (if (and label (string? (list-ref label 1))) (list-ref label 1) ""))
               (width  (if (and width (string->number (list-ref width 1)))
                         (string->number (list-ref width 1))
                         200))
               (height (if (and height (string->number (list-ref height 1)))
                         (string->number (list-ref height 1))
                         200))
               
               (win    (sx-dialog-new #f label width height))
               (table  (sx-hash-make))
              )
              ; parse components of the window
              (lst2component (cddr xml) win table)
              
              ; store the dialog window
              ; the window object
              ; a table related to the window (contains components and callbacks)
              (sx-hash-put! table-dialog (list-ref key 1) (cons win table))
              
              ; return the key to identify the window
              (list-ref key 1)
            )
          )
          'no-key
        )
      )
      ; this is an error
      'no-dialog
    )
  )
  
  ; parses a list of graphical components, widgets
  (define (lst2component lst parent table)
    (cond
      ; empty components
      ((null? lst) #t)
      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      ; if it is a button
      ((equal? (caar lst) 'button)
       (button2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      ; if it is a toggle
      ((equal? (caar lst) 'toggle)
       (toggle2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      ; if it is a text or message
      ((equal? (caar lst) 'message)
       (message2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      ; if it is an edit-box
      ((equal? (caar lst) 'edit-box)
       (editbox2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      
      ; if it is a popup-list
      ((equal? (caar lst) 'popup-list)
       (popup2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      
      ; if it is a list-box
      ((equal? (caar lst) 'list-box)
       (listbox2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      
      ; if it is an image
      ((equal? (caar lst) 'image)
       (image2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      
      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      ; if it is a row
      ((equal? (caar lst) 'row)
       (let ((row (row2component (car lst) parent table)))
         (if (and row (> (length (car lst)) 2))
           (lst2component (cddar lst) row table)
         )
       )
       (lst2component (cdr lst) parent table)
      )
      ; if it is a column
      ((equal? (caar lst) 'column)
       (let ((column (column2component (car lst) parent table)))
         (if (and column (> (length (car lst)) 2))
           (lst2component (cddar lst) column table)
         )
       )
       (lst2component (cdr lst) parent table)
      )
      ; if it is a boxed-column
      ((equal? (caar lst) 'boxed-column)
       (let ((column (boxedcolumn2component (car lst) parent table)))
         (if (and column (> (length (car lst)) 2))
           (lst2component (cddar lst) column table)
         )
       )
       (lst2component (cdr lst) parent table)
      )
      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      ; if it is a tab-column
      ((equal? (caar lst) 'tab-panel)
       (if (sx-dialog? parent)
         (let ((column (tabpanel2component (car lst) parent table)))
           (if (and column (> (length (car lst)) 2))
             (lst2component (cddar lst) column table)
           )
         )
       )
       (lst2component (cdr lst) parent table)
      )
      ; if it is a tab
      ((equal? (caar lst) 'tab)
       (if (sx-tab-panel? parent)
         (let ((column (tab2component (car lst) parent table)))
           (if column
             (let
               ((column (car column))
                (name   (cdr column))
               )
               (sx-tab-panel-append parent name column)
               (if (> (length (car lst)) 2)
                 (lst2component (cddar lst) column table)
               )
             )
           )
         )
       )
       (lst2component (cdr lst) parent table)
      )
      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      ; if it is a spacer
      ((equal? (caar lst) 'spacer)
       (spacer2component (car lst) parent table)
       (lst2component (cdr lst) parent table)
      )
      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      ; if we do not recognize the graphical component, we ignore it
      (else
        (lst2component (cdr lst) parent table)
      )
    )
  )
  
  ; this is the default callback action, for
  ; - button
  ; - toggle
  ; - edit-box
  ; - popup-list
  ; - list-box
  (define (component-callback table key)
    (let ((action (sx-hash-get table key)))
      (if (and action (cdr action) (procedure? (cdr action))) ((cdr action)))
    )
  )
  
  ; ############################################################################
  ;                                  Button
  ; ############################################################################
  
  ; parses and creates a button
  ; it returns nothing
  ; it adds the component to parent and to table
  (define (button2component button parent table)
    (let*
      ((props (cadr button))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (label  (assoc 'label props))
           (width  (assoc 'width props))
           (height (assoc 'height props))
          )
          (let
            ((label  (if (and label  (string? (list-ref label 1)))  
                       (list-ref label 1) 
                       ""))
             (width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (button #f)
            )
            (if (and label (> (string-length label) 0))
              (set! button (sx-button-new parent 
                                          label
                                          (lambda (w e) 
                                            (component-callback table key))))
            )
            (if button
              (begin
                (if width
                  (sx-button-min-width-set! button width)
                )
                (if height
                  (sx-button-min-height-set! button height)
                )
                (sx-hash-put! table key (cons button #f))
              )
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ;                                   Toggle
  ; ############################################################################
  
  ; parses and creates a toggle
  (define (toggle2component toggle parent table)
    (let*
      ((props (cadr toggle))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (label  (assoc 'label props))
           (width  (assoc 'width props))
           (height (assoc 'height props))
           (value  (assoc 'value props))
          )
          (let
            ((label  (if (and label  (string? (list-ref label 1)))  
                       (list-ref label 1) 
                       ""))
             (width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (value  (if (and value 
                              (string? (list-ref value 1))
                              (string=? (list-ref value 1) "1"))
                       #t
                       #f))
             (toggle #f)
            )
            (if (and label (> (string-length label) 0))
              (set! toggle (sx-check-box-new parent 
                                             label 
                                             value
                                             (lambda (w e) 
                                               (component-callback table key))))
            )
            (if toggle
              (begin
                (if width
                  (sx-check-box-min-width-set! toggle width)
                )
                (if height
                  (sx-check-box-min-height-set! toggle height)
                )
                (sx-hash-put! table key (cons toggle #f))
              )
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ;                                 Message
  ; ############################################################################
  
  ; parses and creates a message
  ; it returns nothing
  ; it adds the component to parent and to table
  (define (message2component text parent table)
    (let*
      ((props (cadr text))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (label  (assoc 'label props))
           (width  (assoc 'width props))
           (height (assoc 'height props))
          )
          (let
            ((label  (if (and label  (string? (list-ref label 1)))  
                       (list-ref label 1) 
                       ""))
             (width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (text #f)
            )
            (if (or (and label (> (string-length label) 0)) width)
              (set! text (sx-text-new parent label))
            )
            (if text
              (begin
                (if width
                  (sx-text-min-width-set! text width)
                )
                (if height
                  (sx-text-min-height-set! text height)
                )
                (sx-hash-put! table key (cons text #f))
              )
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ;                               Edit-box
  ; ############################################################################
  
  ; parses and creates a edit-box
  (define (editbox2component editbox parent table)
    (let*
      ((props (cadr editbox))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (label  (assoc 'label props))
           (width  (assoc 'width props))
           (height (assoc 'height props))
           (value  (assoc 'value props))
          )
          (let
            ((label  (if (and label  (string? (list-ref label 1)))  
                       (list-ref label 1) 
                       ""))
             (width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (value  (if (and value 
                              (string? (list-ref value 1)))
                       (list-ref value 1)
                       ""))
             (editbox #f)
            )
            (set! editbox (sx-entry-new parent 
                                        label 
                                        (lambda (w e)
                                          (if (sx-entry-event-enter? w e)
                                            (component-callback table key)))
                                        value))
            (if editbox
              (begin
                (if width
                 (sx-entry-min-width-set! editbox width)
                )
                (if height
                  (sx-entry-min-height-set! editbox height)
                )
                (sx-hash-put! table key (cons editbox #f))
              )
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ;                               Popup-list
  ; ############################################################################
  
  ; parses and creates a popup-list
  (define (popup2component popup parent table)
    (let*
      ((props (cadr popup))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (label  (assoc 'label props))
           (width  (assoc 'width props))
           (height (assoc 'height props))
          )
          (let
            ((label  (if (and label  (string? (list-ref label 1)))  
                       (list-ref label 1) 
                       ""))
             (width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (popup #f)
            )
            (set! popup (sx-choice-new parent 
                                       label 
                                       '("")
                                       (lambda (w e)
                                         (component-callback table key))))
            (if popup
              (begin
                ; Clear any selection
                ; FIXME: this may not be necessary if I can create an empty list
                (sx-choice-clear popup)
                (if width
                  (sx-choice-min-width-set! popup width)
                )
                (if height
                  (sx-choice-min-height-set! popup height)
                )
                (sx-hash-put! table key (cons popup #f))
              )
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ;                                 List-box
  ; ############################################################################
  
  ; parses and creates a list-box
  (define (listbox2component listbox parent table)
    (let*
      ((props (cadr listbox))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (width  (assoc 'width props))
           (height (assoc 'height props))
          )
          (let
            ((width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (listbox #f)
            )
            (set! listbox (sx-list-box-new parent 
                                           #f 
                                           '("")
                                           '(single)
                                           (lambda (w e)
                                             (component-callback table key))))
            (if listbox
              (begin
                (sx-list-box-stretch-x-set! listbox #f)
                (sx-list-box-stretch-y-set! listbox #f)
                ; Clear any selection
                (sx-list-box-clear listbox)
                (if width
                  (sx-list-box-min-width-set! listbox width)
                )
                (if height
                  (sx-list-box-min-height-set! listbox width)
                )
                (sx-hash-put! table key (cons listbox #f))
              )
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ;                               Image
  ; ############################################################################
  
  
  ; parses and creates an image
  ; it returns nothing
  ; it adds the component to parent and to table
  (define (image2component image parent table)
    (let*
      ((props (cadr image))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (width  (assoc 'width props))
           (height (assoc 'height props))
          )
          (let
            ((width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       1))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       1))
             (image #f)
            )
            (set! image (sx-image-canvas parent width height 
                               (lambda ()
                                 (sx-image-canvas-background-draw image)
                                 (sx-image-canvas-bitmap-draw image)
                               ))
            )
            (if image
              (sx-hash-put! table key (cons image #f))
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ;                                    Row
  ; ############################################################################
  
  ; parses and creates a row
  ; it returns the row object
  ; it adds the component to parent and to table
  (define (row2component row parent table)
    (let*
      ((props (cadr row))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (width  (assoc 'width props))
           (height (assoc 'height props))
           (horiz  (assoc 'horizontal-alignment props))
          )
          (let
            ((width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (horiz  (if (and horiz
                              (string? (list-ref horiz 1))
                              (member (list-ref horiz 1) '("center" "left" "right")))
                       (string->symbol (list-ref horiz 1))
                       #f))
             (row #f)
            )
            (set! row (sx-row-new parent))
            (if row
              (begin
                (sx-row-stretch-y-set! row #f)
                (if width
                  (sx-row-min-width-set! row width)
                )
                (if height
                  (sx-row-min-height-set! row height)
                )
                (if horiz
                  (sx-row-alignment-set! row horiz 'center)
                )
                (sx-hash-put! table key (cons row #f))
              )
            )
            ; returns the row object
            row
          )
        )
        #f
      )
    )
  )
  
  ; ############################################################################
  ;                                  Column
  ; ############################################################################
  
  ; parses and creates a column
  ; it returns the column object
  ; it adds the component to parent and to table
  (define (column2component column parent table)
    (let*
      ((props (cadr column))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (width  (assoc 'width props))
           (height (assoc 'height props))
           (vert   (assoc 'vertical-alignment props))
          )
          (let
            ((width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (vert   (if (and vert
                              (string? (list-ref vert 1))
                              (member (list-ref vert 1) '("center" "left" "right")))
                       (string->symbol (list-ref vert 1))
                       #f))
             (column #f)
            )
            (set! column (sx-column-new parent))
            (if column
              (begin
                (sx-column-stretch-x-set! column #f)
                (if width
                  (sx-column-min-width-set! column width)
                )
                (if height
                  (sx-column-min-height-set! column height)
                )
                (if vert
                  (sx-column-alignment-set! column 'center vert)
                )
                (sx-hash-put! table key (cons column #f))
              )
            )
            ; returns the row object
            column
          )
        )
        #f
      )
    )
  )

  
  ; ############################################################################
  ;                                  Boxed-column
  ; ############################################################################
  
  ; parses and creates a boxed-column
  ; it returns the object of the boxed column
  ; it adds the component to parent and to table
  (define (boxedcolumn2component column parent table)
    (let*
      ((props (cadr column))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (label  (assoc 'label props))
           (width  (assoc 'width props))
           (height (assoc 'height props))
           (vert   (assoc 'vertical-alignment props))
          )
          (let
            ((label  (if (and label  (string? (list-ref label 1)))  
                       (list-ref label 1) 
                       ""))
             (width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (vert   (if (and vert
                              (string? (list-ref vert 1))
                              (member (list-ref vert 1) '("center" "left" "right")))
                       (string->symbol (list-ref vert 1))
                       #f))
             (column #f)
            )
            (set! column (sx-boxed-column-new parent label))
            (if column
              (begin
                (sx-boxed-column-stretch-x-set! column #f)
                (if width
                  (sx-boxed-column-min-width-set! column width)
                )
                (if height
                  (sx-boxed-column-min-height-set! column height)
                )
                (if vert
                  (sx-boxed-column-alignment-set! column 'center vert)
                )
                (sx-hash-put! table key (cons column #f))
              )
            )
            ; returns the row object
            column
          )
        )
        #f
      )
    )
  )
  
  ; ############################################################################
  ;                                  Tab-panel
  ; ############################################################################
  
  ; parses and creates a tab-panel
  ; it returns the tab-column object
  ; it adds the component to parent and to table
  (define (tabpanel2component column parent table)
    (let*
      ((props (cadr column))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key    (cadr key))
           (width  (assoc 'width props))
           (height (assoc 'height props))
          )
          (let
            ((width  (if (and width  (string? (list-ref width 1))) 
                       (string->number (list-ref width 1))
                       #f))
             (height (if (and height (string? (list-ref height 1)))
                       (string->number (list-ref height 1))
                       #f))
             (column #f)
            )
            (set! column (sx-tab-panel-new parent))
            (if column
              (begin
                (sx-hash-put! table key (cons column #f))
              )
            )
            ; returns the tab-column object
            column
          )
        )
        #f
      )
    )
  )
  
  ; ############################################################################
  ;                                      Tab
  ; ############################################################################
  
  ; parses and creates a tab
  ; it returns the tab object
  ; it adds the component to parent and to table
  (define (tab2component column parent table)
    (let*
      ((props (cadr column))
       (key   (assoc 'key props))
       (label (assoc 'label props))
      )
      (if (and key 
               (not (sx-hash-get table (cadr key)))
               label
          )
        (let
          ((key   (cadr key))
           (label (cadr label))
          )
          ;(display key)(display " ")(display parent)(newline)
          (let
            ((column #f)
            )
            (set! column (sx-tab-new parent))
            (if column
              (begin
                (sx-hash-put! table key (cons column #f))
              )
            )
            ; returns the tab object and the label
            (cons column label)
          )
        )
        #f
      )
    )
  )
  
  ; ############################################################################
  ;                                 Spacer
  ; ############################################################################
  
  ; parses and creates a spacer
  ; it returns nothing
  ; it adds the component to parent and to table
  (define (spacer2component spacer parent table)
    (let*
      ((props (cadr spacer))
       (key (assoc 'key props))
      )
      (if (and key (not (sx-hash-get table (cadr key))))
        (let
          ((key        (cadr key))
           (width      (assoc 'width props))
           (height     (assoc 'height props))
           (horizontal (assoc 'horizontal props))
           (vertical   (assoc 'vertical props))
          )
          (let
            ((width      (if (and width  (string? (list-ref width 1))) 
                           (string->number (list-ref width 1))
                           #f))
             (height     (if (and height (string? (list-ref height 1)))
                           (string->number (list-ref height 1))
                           #f))
             (horizontal (if (and horizontal 
                                  (string? (list-ref horizontal 1))
                                  (member (list-ref horizontal 1) '("1" "0")))
                           (if (string=? (list-ref horizontal 1) "1")
                             #t
                             #f
                           )
                           #f))
             (vertical   (if (and vertical 
                                  (string? (list-ref vertical 1))
                                  (member (list-ref vertical 1) '("1" "0")))
                           (if (string=? (list-ref vertical 1) "1")
                             #t
                             #f
                           )
                           #f))
             (spacer #f)
            )
            (set! spacer (sx-column-new parent))
            (if spacer
              (begin
                (sx-column-stretch-x-set! spacer horizontal)
                (sx-column-stretch-y-set! spacer vertical)
                (if width
                  (sx-column-min-width-set! spacer width)
                )
                (if height
                  (sx-column-min-height-set! spacer height)
                )
                (sx-hash-put! table key (cons spacer #f))
              )
            )
          )
        )
      )
    )
  )
  
  ; ############################################################################
  ; ############################################################################
  ;                                Functions
  ; ############################################################################
  ; ############################################################################
  
  ; this function shows the dialog window
  (define (sx-show-dialog key)
    (let ((window (sx-hash-get table-dialog key)))
      (if window
        (begin
          (set! opened-dialog (cons key opened-dialog))
          (sx-dialog-show (car window))
        )
      )
    )
  )
  
  ; this function closes down the dialog window
  (define (sx-hide-dialog)
    (if (not (null? opened-dialog))
      (let*
        ((key (car opened-dialog))
         (window (sx-hash-get table-dialog key)))
        (if window
          (begin
            (set! opened-dialog (cdr opened-dialog))
            (sx-dialog-hide (car window))
          )
        )
      )
    )
  )
  
  ; unload a dialog only if it is not opened
  (define (sx-unload-dialog key)
    (if (not (member key opened-dialog))
      (sx-hash-remove! table-dialog key)
    )
  )
  
  ; this function assigns an action to a graphical component
  (define (sx-action-set key-window key-widget function)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if window
        (let*
          ((table-component (cdr window))
           (widget (sx-hash-get table-component key-widget))
          )
          ; FIXME: check that a callback is assigned only when it is possible
          (if widget
            (sx-hash-put! table-component key-widget (cons (car widget) function))
          )
        )
      )
    )
  )
  
  ; for an image widget it specifies the background color
  ; this RGB color is used to fill the bitmap
  (define (sx-image-fill key-window key-widget r g b)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if window
        (let*
          ((table-components (cdr window))
           (widget (sx-hash-get table-components key-widget)))
          (if widget
            (let ((widget (car widget)))
              (sx-image-canvas-background-color widget r g b)
            )
          )
        )
      )
    )
  )
  
  ; for an image widget it specifies an image to display
  (define (sx-image-file key-window key-widget filename)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if window
        (let*
          ((table-components (cdr window))
           (widget (sx-hash-get table-components key-widget)))
          (if widget
            (let ((widget (car widget)))
              (sx-image-canvas-bitmap-set! widget filename)
            )
          )
        )
      )
    )
  )

  ; this function clears the list of items from a popup-list and list-box
  (define (sx-clear-list key-window key-widget)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if window
        (let*
          ((table-components (cdr window))
           (widget (sx-hash-get table-components key-widget)))
          (if widget
            (let ((widget (car widget)))
              (cond
                ((sx-list-box? widget)
                 (sx-list-box-clear widget)
                )
                ((sx-choice? widget)
                 (sx-choice-clear widget)
                )
              )
            )
          )
        )
      )
    )
  )
  
  ; this function adds an item to the list in a popup-list and list-box
  (define (sx-add-list key-window key-widget item)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if window
        (let*
          ((table-components (cdr window))
           (widget (sx-hash-get table-components key-widget)))
          (if widget
            (let ((widget (car widget)))
              (cond
                ((sx-list-box? widget)
                 (sx-list-box-append widget item)
                )
                ((sx-choice? widget)
                 (sx-choice-append widget item)
                )
              )
            )
          )
        )
      )
    )
  )
  
  ; the function sets the label of the graphical component
  ; if it is possible
  ; button
  ; message
  ; toggle
  ; edit-box
  ; popup-list
  ; boxed-column
  (define (sx-label-set key-window key-widget label)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if (and window (string? label))
        (let*
          ((table-widget (cdr window))
           (widget (sx-hash-get table-widget key-widget)))
          (if widget
            (let ((widget (car widget)))
              (cond
                ((sx-button? widget)
                 (sx-button-label-set! widget label)
                )
                ((sx-text? widget)
                 (sx-text-label-set! widget label)
                )
                ((sx-check-box? widget)
                 (sx-check-box-label-set! widget label)
                )
                ((sx-entry? widget)
                 (sx-entry-label-set! widget label)
                )
                ((sx-choice? widget)
                 (sx-choice-label-set! widget label)
                )
                ((sx-boxed-column? widget)
                 (sx-boxed-column-label-set! widget label)
                )
              )
              ; (send widget refresh)
            )
          )
        )
      )
    )
  )
  
  
  ; the function sets the value of the graphical component
  ; toggle - switched on or off
  ; edit-box - text value
  ; popup-list - set selection
  ; listbox - set selection
  (define (sx-value-set key-window key-widget value)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if window
        (let*
          ((table-widget (cdr window))
           (widget (sx-hash-get table-widget key-widget)))
          (if widget
            (let ((widget (car widget)))
              (cond
                ((sx-entry? widget)
                 (if (string? value)
                   (sx-entry-text-set! widget value)
                 )
                )
                ((sx-check-box? widget)
                 (if (member value '(#f #t))
                   (sx-check-box-value-set! widget value)
                 )
                )
                ((sx-choice? widget)
                 (if (and (number? value) 
                          (integer? value) 
                          (exact? value)
                          (<= 0 value (- (sx-choice-get-number widget) 1))
                     )
                   (sx-choice-set-selection widget value)
                 )
                )
                ((sx-list-box? widget)
                 (if (and (number? value) 
                          (integer? value) 
                          (exact? value)
                          (<= 0 value (- (sx-list-box-get-number widget) 1))
                     )
                   (sx-list-box-set-selection widget value)
                 )
                )
                ((sx-tab-panel? widget)
                 (if (and (number? value) 
                          (integer? value) 
                          (exact? value)
                          (<= 0 value (- (sx-tab-panel-get-number widget) 1))
                     )
                   (sx-tab-panel-set-selection widget value)
                 )
                )
              )
            )
          )
        )
      )
    )
  )
  
  ; the function returns the value of the graphical component
  ; toggle - switched on or off
  ; edit-box - text value
  ; popup-list - get selected item index
  ; listbox - get selected item index
  (define (sx-value-get key-window key-widget)
    (let ((window (sx-hash-get table-dialog key-window)))
      (if window
        (let*
          ((table-widget (cdr window))
           (widget (sx-hash-get table-widget key-widget)))
          (if widget
            (let ((widget (car widget)))
              (cond
                ((sx-entry? widget)
                 (sx-entry-text-get widget)
                )
                ((sx-check-box? widget)
                 (sx-check-box-value-get widget)
                )
                ((sx-choice? widget)
                 (let ((index (sx-choice-get-selection widget)))
                   ; ensure that a number is returned all the time
                   (if (not index)
                     0
                     index
                   )
                 )
                )
                ((sx-list-box? widget)
                 (let ((index (sx-list-box-get-selection widget)))
                   ; ensure that a number is returned all the time
                   (if (not index)
                     0
                     index
                   )
                 )
                )
              )
            )
          )
        )
      )
    )
  )
    
) ; end of module


